/*
 * $Id: diald-top.c,v 0.0.1.4 1996/01/12 03:49:13 gjhurlbu Exp $
 *
 * DialD Packet Statistics Program
 * (c) 1995 Gavin J. Hurlbut
 * gjhurlbu@beirdo.uplink.on.ca
 *
 * $Log: diald-top.c,v $
 * Revision 0.0.1.4  1996/01/12  03:49:13  gjhurlbu
 * Minor revisions from 0.0.1.3
 *
 * Revision 0.0.1.3  1996/01/02  03:37:59  gjhurlbu
 * Blindly trying Jeff Epler's second patch
 *
 * Revision 0.0.1.1  1995/11/13  04:05:24  gjhurlbu
 * Added Redraw to the help file
 * Ready for the release again....
 *
 * Revision 0.0.0.6  1995/11/12  23:50:25  gjhurlbu
 * Added Help screen
 * Redraw (^R) now works
 *
 * Revision 0.0.0.5  1995/11/07  04:22:34  gjhurlbu
 * Tried to fix the stupid bug (when you hit c then start typing)
 * Also attempted to add a redraw -- doesn't work yet
 *
 * Revision 0.0.0.4  1995/11/07  03:50:46  gjhurlbu
 * Added FIFO command interface
 * Still needs a help page
 *
 * Revision 0.0.0.3  1995/11/06  06:53:37  gjhurlbu
 * Added Link Status:
 *
 * Revision 0.0.0.2  1995/11/05  18:35:07  gjhurlbu
 * Started to modularize
 * Added -c and -m command line parms
 *
 * Revision 0.0.0.1  1995/11/01  08:10:04  gjhurlbu
 * Out to full Major.Minor.Patch.Hack revisions.
 * I should have done this earlier.
 *
 * Revision 0.0  1995/11/01  08:08:03  gjhurlbu
 * Starting Again at Rev 0.0 to allow for proper version numbering
 * outdated all older versions
 *
 */

/* Included Header Files */
#include <stdio.h>
#include <unistd.h>
#include <ncurses.h>
#include <string.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <stdlib.h>

char *strdup(const char*);

/* Global Definitions */
#define NUM_CMDS	8
#define HEAD_LINES	7
#define REV		"$Revision: 0.0.1.4 $"

#define HASHSIZE 65536

struct hostrec {
	struct in_addr addr;
	char *name;
	struct hostrec *next;
} *hosthash[HASHSIZE];

/* Protoypes */
int main(int argc, char **argv);
int fifo_command(char *command);	/* Send commands to diald's FIFO */
int read_command_line(int argc, char **argv);
void display_usage(int argc, char **argv);
char *strstrip(char *line);
void validate_command(char *command);
void get_command(WINDOW *screen, char *command);
void helpscreen(void);
const char *resolve(const char *aaddr);
const char *service(int port, const char *proto);
int hash(struct in_addr addr);
struct hostrec *gettab(struct in_addr addr);
int probetab(struct in_addr addr);

/* Global Variables */
char control[256];
char monitor[256];
char cmd[NUM_CMDS+1][10]={"up","down","force","unforce","block","unblock",			"quit","delay-quit",""};
char curr_state[1024];
int numeric;
WINDOW *full, *sub;



int main(int argc, char **argv)
{
	FILE *fp_monitor;
	char key, line[1024];
	int count, queue, state, quit;

	read_command_line(argc,argv);

	sprintf(line,"monitor %s",monitor);
	fifo_command(line);

	if ((fp_monitor=fopen(monitor,"rt"))==NULL)
	{
		fprintf(stderr,"Could not open diald monitor pipe\n");
		exit(1);
	}


	quit = queue = count = state = 0;

	initscr();
	full=newwin(0,0,0,0);
	sub=subwin(full,LINES-HEAD_LINES,COLS,HEAD_LINES,0);
	curs_set(0);
	nodelay(stdscr,TRUE);
	cbreak();
	noecho();

	strcpy(curr_state,"");
	werase(full);
	mvwprintw(full,0,0,"DialD Packet statistics program (c) 1995 Gavin J. Hurlbut");
	mvwprintw(full,1,0,"%s \tgjhurlbu@beirdo.uplink.on.ca\t(h) for help",REV);
	mvwprintw(full,HEAD_LINES-2,0,"Type       Destination                       Source                   TTL");
	mvwprintw(full,HEAD_LINES-1,0,"===============================================================================");

	wrefresh(full);

	while ( !quit )
	{
		fgets(line,1024,fp_monitor);
		if ( feof(fp_monitor) ) break;
		
		if ( strncmp("QUEUE",line,5) == 0 )
		{
			queue=1;
			count=0;
			werase(sub);
		}
		else if ( strncmp("END QUEUE",line,9) == 0 )
		{
			wmove(full,HEAD_LINES-4,0);
			wclrtoeol(full);

			sprintf(line,"Link Status:  %s          ",curr_state);
			mvwprintw(full,HEAD_LINES-4,0,"%s",line);

			sprintf(line,"Packets Queued:  %d",count);
			mvwprintw(full,HEAD_LINES-4,(COLS-strlen(line))/2,"%s",line);
			if ( count == 0 )
			{
				mvwprintw(sub,0,5,"No packets queued\n");
			}
			queue=0;
			touchwin(full); 
			wrefresh(full);
		}
		else if ( queue==1 )
		{
 			if ( count <= LINES-HEAD_LINES )
			{
				char proto[6];
				char destip[16];
				int  destport;
				char *destports;
				char sourceip[16];
				int  sourceport;
				char *sourceports;
				char ttl[10];
				sscanf(line,"%s %[0123456789.]/%d %[0123456789.]/%d %s",proto,destip,&destport,sourceip,&sourceport,ttl);
				destports=(char *)service(destport,proto);
				sourceports=(char *)service(sourceport,proto);
				mvwprintw(sub,count++,0, "%-4.4s %22.22s/%-8.8s %22.22s/%-8.8s  %s", proto, resolve(destip), destports, resolve(sourceip), sourceports, ttl);
				free(destports);
				free(sourceports);
				/* mvwprintw(sub,count++,0,"%s",line); */
			}
			else
			{
				count++;
			}
		}
		else if ( strncmp( "STATE",line,5) == 0 )   state=1;
		else if ( state == 1)
		{
			state=0;
			strcpy(curr_state,strstrip(line));

			wmove(full,HEAD_LINES-4,0);
			wclrtoeol(full);

			sprintf(line,"Link Status:  %s          ",curr_state);
			mvwprintw(full,HEAD_LINES-4,0,"%s",line);

			sprintf(line,"   Packets Queued:  %d   ",count);
			mvwprintw(full,HEAD_LINES-4,(COLS-strlen(line))/2,"%s",line);
			touchwin(full); 
			wrefresh(full);
		}

		key=getch();

		switch(key) 
		{
		case 'q':
			quit=1;
			break;
		case 'h':
			helpscreen();
			touchwin(full);
			wrefresh(full);
			break;
		case 'c':
			get_command(full,line);
			validate_command(line);
			fifo_command(line);
			wmove(full,4,0);
			wclrtoeol(full);
			wrefresh(full);
			break;
		case 'n':
			numeric=1-numeric;
			break;
		case 18:
			wrefresh(curscr);
			break;
		}
	}

	fclose(fp_monitor);
	
	fifo_command("monitor");

	nodelay(stdscr,FALSE);
	nocbreak();
	curs_set(1);
	delwin(sub);
	delwin(full);
	endwin();
}
void validate_command(char *command)
{
	int i, len;

	for(i=0;i<NUM_CMDS;i++)
	{
		len=strlen(cmd[i]);
		if ( strncasecmp(cmd[i],strstrip(command),len) == 0 ) break;
	}

	strcpy(command,cmd[i]);
}

void get_command(WINDOW *screen, char *command)
{
	mvwprintw(screen,4,0,"FIFO Command:");
	touchwin(screen);
	wrefresh(screen);
	curs_set(1);
	echo();
	mvwgetstr(screen,4,15,command);
	noecho();
	curs_set(0);
}

void helpscreen(void)
{
	werase(sub);
	mvwprintw(sub,1,0,"(q)\tQuit diald-top");
	mvwprintw(sub,2,0,"(h)\tDisplay Help Screen (this screen)");
	mvwprintw(sub,3,0,"(c)\tSend a command to the diald control FIFO");
	mvwprintw(sub,4,8,"Valid Commands:");
	mvwprintw(sub,5,10,"up           Bring the link up");
	mvwprintw(sub,6,10,"down         Bring the link down");
	mvwprintw(sub,7,10,"force        Force the link up until further notice");
	mvwprintw(sub,8,10,"unforce      No longer force the link up");
	mvwprintw(sub,9,10,"block        Force the link down until further notice");
	mvwprintw(sub,10,10,"unblock      No longer force the link down");
	mvwprintw(sub,11,10,"quit         Forcibly quit diald (and shut down the link)");
	mvwprintw(sub,12,10,"delay-quit   Quit diald the next time the link comes down");

	mvwprintw(sub,13,0,"(^R)\tRedraw screen");
	mvwprintw(sub,14,0,"(n)\tToggle Numeric IP vs. Resolved Names");
	mvwprintw(sub,16,26,"Press any key to resume");

	touchwin(full);
	wrefresh(full);

	while(getch()==ERR);
}

int read_command_line(int argc, char **argv)
{
	extern char *optarg;
	int opt;

	strcpy(monitor,DIALD_MON);
	strcpy(control,DIALD_CTL);

	numeric=0;

	while ( (opt=getopt(argc,argv,"nhc:m:")) != -1 )
	{
		switch(opt) {
		case 'n':
			numeric=1;
			break;
		case 'c':
			strcpy(control,optarg);
			break;
		case 'm':
			strcpy(monitor,optarg);
			break;
		case ':':
		case '?':
		case 'h':
			display_usage(argc,argv);
			break;
		}
	}
	return(0);
}

int fifo_command(char *command)
{
	FILE *fp_control;

	if (strcmp("",command)==0)  return(1);	/* Ignore NULL commands */
	if ((fp_control=fopen(control,"at"))==NULL)
	{
		fprintf(stderr,"Could not open diald control pipe\n");
		exit(1);
	}
	fprintf(fp_control,"%s\n",command);
	fclose(fp_control);
	return(0);
}

void display_usage(int argc, char **argv)
{
	fprintf(stderr,"\nDialD Packet statistics program (c) 1995 Gavin J. Hurlbut\n");
	fprintf(stderr,"%s \tgjhurlbu@beirdo.uplink.on.ca\n\n",REV);
	fprintf(stderr,"Proper usage:\n\n");
	fprintf(stderr,"\t%s [-n] [-c control_FIFO] [-m monitor_FIFO]\n\n",argv[0]);
	fprintf(stderr,"If the -n option is used, the IP addresses will not be resolved\n");
	fprintf(stderr,"  otherwise, all IP addresses will be resolved (when the link is up)\n");
	fprintf(stderr,"The -c and -m options are optional.  If they are not included, the files\n");
	fprintf(stderr,"will default to:\n\n");
	fprintf(stderr,"\tControl FIFO: %s\n",DIALD_CTL);
	fprintf(stderr,"\tMonitor FIFO: %s\n\n\n",DIALD_MON);
	
	exit(2);
}

char *strstrip(char *line)
{
	int len, i;

	len=strlen(line);
	for(i=0;i<len && !isspace(line[i]);i++);
	line[i]='\0';

	return(line);
}

const char *resolve(const char *aaddr) {
	struct in_addr addr;
	if(inet_aton(aaddr,&addr)==0 || numeric) return aaddr;
	if(probetab(addr)) return gettab(addr)->name;
	if(!strcmp(curr_state,"UP")) {
		struct hostent *host;
		struct hostrec *hr;
		host=gethostbyaddr((char*)&addr,sizeof(addr),AF_INET);
		if(host==NULL) {
			if(h_errno==HOST_NOT_FOUND||h_errno==NO_ADDRESS) {
				hr=gettab(addr);
				hr->name=strdup(aaddr);
				hr->addr=addr;
				return hr->name;
			} else return aaddr;
		}
		hr=gettab(addr);
		hr->name=strdup(host->h_name);
		hr->addr=addr;
		return hr->name;
	}
	return aaddr;
}

int hash(struct in_addr addr) {
	unsigned char *c=(unsigned char*)&addr;
	return (c[0]^c[2])+256*(c[1]^c[3]);
}
		
struct hostrec *gettab(struct in_addr addr) {
	int h=hash(addr);
	struct hostrec *hr=hosthash[h];
	while(hr) {
		if(!memcmp(&addr,&(hr->addr),sizeof(struct in_addr))) return hr;
		hr=hr->next;
	}
	hr=malloc(sizeof(struct hostrec));
	hr->next=hosthash[h];
	hosthash[h]=hr;
	return hr;
}
		
int probetab(struct in_addr addr) {
	struct hostrec *hr=hosthash[hash(addr)];
	while(hr) {
		if(!memcmp(&addr,&(hr->addr),sizeof(struct in_addr))) return 1;
		hr=hr->next;
	}
	return 0;
}

const char *service(int port, const char *proto) {
	char *sstr;
	struct servent *serv=(setservent(1),getservbyport(htons(port),proto));
	if(serv==NULL) { sstr=malloc(6); sprintf(sstr,"%-5d",port); }
	else { sstr=strdup(serv->s_name); }
	return sstr;
}
	
